Attribute VB_Name = "Geometry"
Option Explicit

' Find the polygon's centroid.
Public Sub FindCentroid(ByRef X As Single, ByRef Y As Single)
Dim pt As Integer
Dim second_factor As Single
Dim polygon_area As Single

    ' Add the first point at the end of the array.
    ReDim Preserve m_Points(1 To m_NumPoints + 1)
    m_Points(m_NumPoints + 1) = m_Points(1)

    ' Find the centroid.
    X = 0
    Y = 0
    For pt = 1 To m_NumPoints
        second_factor = _
            m_Points(pt).X * m_Points(pt + 1).Y - _
            m_Points(pt + 1).X * m_Points(pt).Y
        X = X + (m_Points(pt).X + m_Points(pt + 1).X) * _
            second_factor
        Y = Y + (m_Points(pt).Y + m_Points(pt + 1).Y) * _
            second_factor
    Next pt

    ' Divide by 6 times the polygon's area.
    polygon_area = PolygonArea
    X = X / 6 / polygon_area
    Y = Y / 6 / polygon_area

    ' If the values are negative, the polygon is
    ' oriented counterclockwise. Reverse the signs.
    If X < 0 Then
        X = -X
        Y = -Y
    End If
End Sub

' Find the indexes of three points that form an ear.
Private Sub FindEar(ByRef A As Integer, ByRef B As Integer, ByRef C As Integer)
    For A = 1 To m_NumCopyPoints
        If FormsEar(A, A + 1, A + 2) Then
            B = A + 1
            If B > m_NumCopyPoints Then B = 1
            C = B + 1
            If C > m_NumCopyPoints Then C = 1
            Exit Sub
        End If
    Next A

    ' We should never get here because there should
    ' always be at least two ears.
    Stop
End Sub
' Return True if the three points form an ear.
Private Function FormsEar(ByVal A As Integer, ByVal B As Integer, ByVal C As Integer) As Boolean
Dim pt As Integer
Dim test_points(1 To 3) As POINTAPI

    ' Assume the points form an ear.
    FormsEar = True

    ' See if the angle ABC is concave.
    If GetAngle( _
        m_CopyPoints(A).X, m_CopyPoints(A).Y, _
        m_CopyPoints(B).X, m_CopyPoints(B).Y, _
        m_CopyPoints(C).X, m_CopyPoints(C).Y) > 0 _
    Then
        ' This is a concave corner so the triangle
        ' cannot be an ear.
        FormsEar = False
        Exit Function
    End If

    ' Make the triangle A, B, C.
    test_points(1) = m_CopyPoints(A)
    test_points(2) = m_CopyPoints(B)
    test_points(3) = m_CopyPoints(C)

    ' Check the other points to see if they lie in the
    ' triangle A, B, C.
    For pt = 1 To m_NumCopyPoints
        If pt <> A And pt <> B And pt <> C Then
            With m_CopyPoints(pt)
                If PointInPolygon(.X, .Y, 3, test_points) Then
                    ' This point is in the triangle so
                    ' this is not an ear.
                    FormsEar = False
                    Exit For
                End If
            End With
        End If
    Next pt
End Function
' If the polygon is oriented counterclockwise,
' reverse the order of its points.
Private Sub OrientPolygonClockwise(ByVal num_polygon_points As Integer, polygon_points() As POINTAPI)
Dim pt As Integer
Dim temp_point As POINTAPI

    If Not PolygonIsOrientedClockwise() Then
        For pt = 1 To num_polygon_points \ 2
            temp_point = polygon_points(pt)
            polygon_points(pt) = polygon_points(num_polygon_points - pt + 1)
            polygon_points(num_polygon_points - pt + 1) = temp_point
        Next pt
    End If
End Sub


' Return True if the point is in the polygon.
Public Function PointInPolygon(ByVal X As Single, ByVal Y As Single, ByVal num_polygon_points As Integer, polygon_points() As POINTAPI) As Boolean
Dim pt As Integer
Dim total_angle As Single

    ' Get the angle between the point and the
    ' first and last vertices.
    total_angle = GetAngle( _
        polygon_points(num_polygon_points).X, polygon_points(num_polygon_points).Y, _
        X, Y, _
        polygon_points(1).X, polygon_points(1).Y)

    ' Add the angles from the point to each other
    ' pair of vertices.
    For pt = 1 To num_polygon_points - 1
        total_angle = total_angle + GetAngle( _
            polygon_points(pt).X, polygon_points(pt).Y, _
            X, Y, _
            polygon_points(pt + 1).X, polygon_points(pt + 1).Y)
    Next pt

    ' The total angle should be 2 * PI or -2 * PI if
    ' the point is in the polygon and close to zero
    ' if the point is outside the polygon.
    PointInPolygon = (Abs(total_angle) > 0.000001)
End Function
' Return True if the polygon is oriented clockwise.
Public Function PolygonIsOrientedClockwise() As Boolean
    PolygonIsOrientedClockwise = SignedPolygonArea() < 0
End Function

' Return the polygon's area in "square" pixels.
' Add the areas of the trapezoids defined by the
' polygon's edges dropped to the X-axis. When the
' program considers a bottom edge of a polygon, the
' calculation gives a negative area so the space
' between the polygon and the axis is subtracted,
' leaving the polygon's area. This method gives odd
' results for non-simple polygons.
Public Function PolygonArea() As Single
    ' Return the absolute value of the signed area.
    ' The signed area is negative if the polyogn is
    ' oriented clockwise.
    PolygonArea = Abs(SignedPolygonArea())
End Function

' Return True if the polygon is convex.
Public Function PolygonIsConvex() As Boolean
Dim pt As Integer
Dim cross_product As Single
Dim got_negative As Boolean
Dim got_positive As Boolean

    ' For each set of three adjacent points A, B, C,
    ' find the dot product AB  BC. If the sign of
    ' all the dot products is the same, the angles
    ' are all positive or negative (depending on the
    ' order in which we visit them) so the polygon
    ' is convex.

    ' Assume the polygon is non-convex.
    PolygonIsConvex = False

    ' Look at the first set of points.
    cross_product = CrossProductLength( _
        m_Points(m_NumPoints - 1).X, m_Points(m_NumPoints - 1).Y, _
        m_Points(m_NumPoints).X, m_Points(m_NumPoints).Y, _
        m_Points(1).X, m_Points(1).Y)
    If cross_product < 0 Then
        got_negative = True
    ElseIf cross_product > 0 Then
        got_positive = True
    End If

    ' Look at the second set of points.
    cross_product = CrossProductLength( _
        m_Points(m_NumPoints).X, m_Points(m_NumPoints).Y, _
        m_Points(1).X, m_Points(1).Y, _
        m_Points(2).X, m_Points(2).Y)
    If cross_product < 0 Then
        got_negative = True
    ElseIf cross_product > 0 Then
        got_positive = True
    End If
    If got_negative And got_positive Then Exit Function

    ' Look at the remaining triples of points.
    For pt = 1 To m_NumPoints - 2
        cross_product = CrossProductLength( _
            m_Points(pt).X, m_Points(pt).Y, _
            m_Points(pt + 1).X, m_Points(pt + 1).Y, _
            m_Points(pt + 2).X, m_Points(pt + 2).Y)
        If cross_product < 0 Then
            got_negative = True
        ElseIf cross_product > 0 Then
            got_positive = True
        End If
        If got_negative And got_positive Then Exit Function
    Next pt

    ' If we got this far, the polygon is convex.
    PolygonIsConvex = True
End Function

' Return the angle with tangent opp/hyp. The returned
' value is between PI and -PI.
Public Function ATan2(ByVal opp As Single, ByVal adj As Single) As Single
Dim angle As Single

    ' Get the basic angle.
    If Abs(adj) < 0.0001 Then
        angle = PI / 2
    Else
        angle = Abs(Atn(opp / adj))
    End If

    ' See if we are in quadrant 2 or 3.
    If adj < 0 Then
        ' angle > PI/2 or angle < -PI/2.
        angle = PI - angle
    End If

    ' See if we are in quadrant 3 or 4.
    If opp < 0 Then
        angle = -angle
    End If

    ' Return the result.
    ATan2 = angle
End Function
' Return the cross product AB x BC.
' The cross product is a vector perpendicular to AB
' and BC having length |AB| * |BC| * Sin(theta) and
' with direction given by the right-hand rule.
' For two vectors in the X-Y plane, the result is a
' vector with X and Y components 0 so the Z component
' gives the vector's length and direction.
Public Function CrossProductLength( _
    ByVal Ax As Single, ByVal Ay As Single, _
    ByVal Bx As Single, ByVal By As Single, _
    ByVal Cx As Single, ByVal Cy As Single _
  ) As Single
Dim BAx As Single
Dim BAy As Single
Dim BCx As Single
Dim BCy As Single

    ' Get the vectors' coordinates.
    BAx = Ax - Bx
    BAy = Ay - By
    BCx = Cx - Bx
    BCy = Cy - By

    ' Calculate the Z coordinate of the cross product.
    CrossProductLength = BAx * BCy - BAy * BCx
End Function
' Return the dot product AB  BC.
' Note that AB  BC = |AB| * |BC| * Cos(theta).
Private Function DotProduct( _
    ByVal Ax As Single, ByVal Ay As Single, _
    ByVal Bx As Single, ByVal By As Single, _
    ByVal Cx As Single, ByVal Cy As Single _
  ) As Single
Dim BAx As Single
Dim BAy As Single
Dim BCx As Single
Dim BCy As Single

    ' Get the vectors' coordinates.
    BAx = Ax - Bx
    BAy = Ay - By
    BCx = Cx - Bx
    BCy = Cy - By

    ' Calculate the dot product.
    DotProduct = BAx * BCx + BAy * BCy
End Function

' Return the angle ABC.
' Return a value between PI and -PI.
' Note that the value is the opposite of what you might
' expect because Y coordinates increase downward.
Public Function GetAngle(ByVal Ax As Single, ByVal Ay As Single, ByVal Bx As Single, ByVal By As Single, ByVal Cx As Single, ByVal Cy As Single) As Single
Dim dot_product As Single
Dim cross_product As Single

    ' Get the dot product and cross product.
    dot_product = DotProduct(Ax, Ay, Bx, By, Cx, Cy)
    cross_product = CrossProductLength(Ax, Ay, Bx, By, Cx, Cy)

    ' Calculate the angle.
    GetAngle = ATan2(cross_product, dot_product)
End Function

' Remove an ear from the polygon.
Private Sub RemoveEar()
Dim A As Integer
Dim B As Integer
Dim C As Integer
Dim old_point As Integer
Dim new_point As Integer

    ' Find an ear.
    FindEar A, B, C

    ' Create a new triangle for the ear.
    m_NumTriangles = m_NumTriangles + 1
    With m_Triangles(m_NumTriangles)
        .Points(1) = m_CopyPoints(A)
        .Points(2) = m_CopyPoints(B)
        .Points(3) = m_CopyPoints(C)
    End With

    ' Remove the ear from the polygon.
    new_point = 0
    For old_point = 1 To m_NumCopyPoints
        If old_point <> B Then
            new_point = new_point + 1
            m_CopyPoints(new_point) = m_CopyPoints(old_point)
        End If
    Next old_point
    m_NumCopyPoints = m_NumCopyPoints - 1

    ' Repeat the first two points.
    m_CopyPoints(m_NumCopyPoints + 1) = m_CopyPoints(1)
    m_CopyPoints(m_NumCopyPoints + 2) = m_CopyPoints(2)
End Sub
' Return the polygon's area in "square" pixels.
' Add the areas of the trapezoids defined by the
' polygon's edges dropped to the X-axis. When the
' program considers a bottom edge of a polygon, the
' calculation gives a negative area so the space
' between the polygon and the axis is subtracted,
' leaving the polygon's area. This method gives odd
' results for non-simple polygons.
'
' The value will be negative if the polyogn is
' oriented clockwise.
Private Function SignedPolygonArea() As Single
Dim pt As Integer
Dim area As Single

    ' Add the first point to the end.
    ReDim Preserve m_Points(1 To m_NumPoints + 1)
    m_Points(m_NumPoints + 1) = m_Points(1)

    ' Get the areas.
    For pt = 1 To m_NumPoints
        area = area + _
            (m_Points(pt + 1).X - m_Points(pt).X) * _
            (m_Points(pt + 1).Y + m_Points(pt).Y) / 2
    Next pt

    ' Return the result.
    SignedPolygonArea = area
End Function

' Traingulate the polygon.
'
' For a nice, detailed explanation of this method,
' see Ian Garton's Web page:
' http://www-cgrl.cs.mcgill.ca/~godfried/teaching/cg-projects/97/Ian/cutting_ears.html
Public Sub Triangulate()
Dim pt As Integer

    ' Copy the points into a new array.
    m_NumCopyPoints = m_NumPoints
    ReDim m_CopyPoints(1 To m_NumCopyPoints + 2)
    For pt = 1 To m_NumPoints
        m_CopyPoints(pt) = m_Points(pt)
    Next pt

    ' Orient the polygon.
    OrientPolygonClockwise m_NumCopyPoints, m_CopyPoints

    ' Repeat the first two points.
    m_CopyPoints(m_NumCopyPoints + 1) = m_Points(1)
    m_CopyPoints(m_NumCopyPoints + 2) = m_Points(2)

'@
For pt = 1 To m_NumCopyPoints
    With m_CopyPoints(pt)
        EditorForm.CurrentX = .X
        EditorForm.CurrentY = .Y
        EditorForm.Print pt
    End With
Next pt
'@

    ' Make room for the triangles.
    m_NumTriangles = 0
    ReDim m_Triangles(1 To m_NumPoints - 2)

    ' While the copy of the polygon has more than
    ' three points, remove an ear.
    Do While m_NumCopyPoints > 3
        ' Remove an ear from the polygon.
        RemoveEar
    Loop

    ' Copy the last three points into their own triangle.
    m_NumTriangles = m_NumTriangles + 1
    With m_Triangles(m_NumTriangles)
        For pt = 1 To 3
            .Points(pt) = m_CopyPoints(pt)
        Next pt
    End With
End Sub

